@@ -1,14 +1,10 @@
 use crate::bounds::Bounds1;
 use crate::constellation::Constellation;
 use crate::star_node::StarNodeType;
+use crate::util::gaussian_logp;
 use crate::NNVFloat;
 use log::{debug, info};
-use ndarray::concatenate;
-use ndarray::s;
-use ndarray::Array1;
-use ndarray::Axis;
-use ndarray::Dimension;
-use ndarray::Ix2;
+use ndarray::{concatenate, s, Array1, Axis, Dimension, Ix2};
 use rand::distributions::{Bernoulli, Distribution};
 use rand::Rng;
 use std::iter;
@@ -72,6 +68,7 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
         cdf_samples: usize,
         max_iters: usize,
         time_limit_opt: Option<Duration>,
+        stability_eps: T,
     ) -> Option<(Vec<Array1<T>>, T, T)> {
         let start_time = Instant::now();
         let mut best_sample = None;
@@ -87,16 +84,16 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
                                 total_infeasible_cdf: &mut T,
                                 rng: &mut R|
          -> usize {
-            info!("Infeasible reset!");
+            info!("Infeasible reset at depth {}!", path.len());
             me.set_feasible(x, false);
-            let mut infeas_cdf = me
-                .constellation
-                .get_node_cdf(x, cdf_samples, max_iters, rng);
+            let mut infeas_cdf =
+                me.constellation
+                    .get_node_cdf(x, cdf_samples, max_iters, rng, stability_eps);
             path.drain(..).rev().for_each(|x| {
                 // check if all chilren are infeasible
                 if me
                     .constellation
-                    .get_node_child_ids(x)
+                    .get_node_child_ids(x, stability_eps)
                     .iter()
                     .any(|&x| me.get_feasible(x))
                 {
@@ -105,9 +102,13 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
                         .add_node_cdf(x, T::neg(T::one()) * infeas_cdf);
                 } else {
                     me.set_feasible(x, false);
-                    infeas_cdf = me
-                        .constellation
-                        .get_node_cdf(x, cdf_samples, max_iters, rng)
+                    infeas_cdf = me.constellation.get_node_cdf(
+                        x,
+                        cdf_samples,
+                        max_iters,
+                        rng,
+                        stability_eps,
+                    );
                 }
             });
             *total_infeasible_cdf += infeas_cdf;
@@ -122,7 +123,8 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
                 return None;
             }
             // Try to sample and return if a valid sample is found
-            let safe_sample_opt = self.try_safe_sample(current_node, rng, num_samples, max_iters);
+            let safe_sample_opt =
+                self.try_safe_sample(current_node, rng, num_samples, max_iters, stability_eps);
             debug_assert!(self
                 .constellation
                 .try_get_gaussian_distribution(current_node)
@@ -130,25 +132,28 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
             match safe_sample_opt {
                 Ok((safe_sample, est_cost)) => {
                     info!(
-                        "Safe sample with value {} less than {} at depth {:?}",
+                        "Safe sample with value {} less than {} at depth {}",
                         est_cost,
                         self.safe_value,
                         path.len()
                     );
                     return Some((vec![safe_sample], path_logp, total_infeasible_cdf));
                 }
-                Err((unsafe_sample, val)) => {
-                    best_sample = Some(unsafe_sample);
-                    best_sample_val = val;
+                Err(Some((unsafe_sample, val))) => {
+                    if val < best_sample_val {
+                        best_sample = Some(unsafe_sample);
+                        best_sample_val = val;
+                    }
                 }
+                Err(None) => {}
             };
             // check feasibility of current node
             {
                 // makes the assumption that bounds are on 0th dimension of output
-                let output_bounds = if current_node != 0 {
-                    self.constellation.get_node_output_bounds(current_node)
-                } else {
+                let output_bounds = if current_node == 0 {
                     (T::neg_infinity(), T::infinity())
+                } else {
+                    self.constellation.get_node_output_bounds(current_node)
                 };
                 debug!("Output bounds: {:?}", output_bounds);
                 debug_assert!(self
@@ -168,6 +173,7 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
                         rng,
                         num_samples,
                         max_iters,
+                        stability_eps,
                     );
                     return Some((safe_sample, path_logp, total_infeasible_cdf));
                 } else if output_bounds.0 > self.safe_value {
@@ -182,7 +188,10 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
                         rng,
                     );
                     continue;
-                } else if let StarNodeType::Leaf = self.constellation.get_node_type(current_node) {
+                } else if let StarNodeType::Leaf = self
+                    .constellation
+                    .get_node_type(current_node, stability_eps)
+                {
                     // do procedure to select safe part
                     info!(
                         "Safe sample with value at most {} at leaf (bounds {:?})",
@@ -194,6 +203,7 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
                         num_samples,
                         max_iters,
                         self.safe_value,
+                        stability_eps,
                     );
                     return Some((safe_sample, path_logp, total_infeasible_cdf));
                 }
@@ -204,7 +214,7 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
             if let Some(time_limit) = time_limit_opt {
                 if start_time.elapsed() >= time_limit {
                     info!(
-                        "Unsafe sample after timeout with value with value {:?} at depth {}",
+                        "Unsafe sample after timeout with value {:?} at depth {}",
                         best_sample_val,
                         path.len()
                     );
@@ -217,8 +227,14 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
                 .is_some());
             // expand node
             {
-                let result =
-                    self.select_safe_child(current_node, path_logp, cdf_samples, max_iters, rng);
+                let result = self.select_safe_child(
+                    current_node,
+                    path_logp,
+                    cdf_samples,
+                    max_iters,
+                    rng,
+                    stability_eps,
+                );
                 match result {
                     Some((node, logp)) => {
                         debug_assert!(node != current_node);
@@ -246,29 +262,65 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
         rng: &mut R,
         num_samples: usize,
         max_iters: usize,
-    ) -> Result<(Array1<T>, T), (Array1<T>, T)> {
-        let unsafe_sample =
-            self.constellation
-                .sample_gaussian_node(current_node, rng, num_samples, max_iters);
+        stability_eps: T,
+    ) -> Result<(Array1<T>, T), Option<(Array1<T>, T)>> {
+        let unsafe_sample = self.constellation.sample_gaussian_node(
+            current_node,
+            rng,
+            num_samples,
+            max_iters,
+            stability_eps,
+        );
+        let input_set = self
+            .constellation
+            .get_node_reduced_input_polytope(current_node);
         let unsafe_len = unsafe_sample[0].len();
-        let fixed_input_part: Option<Array1<T>> = {
-            self.constellation
-                .get_input_bounds()
-                .as_ref()
-                .map(|bounds| {
-                    let fixed_bounds: Bounds1<T> = bounds.split_at(unsafe_len).1;
-                    let fixed_array: Array1<T> = fixed_bounds.lower().to_owned();
-                    fixed_array
-                })
-        };
-        let dnn_idx = self.constellation.get_node_dnn_index(current_node);
-        let mut best_sample = Array1::zeros(1);
+        let fixed_input_part: Option<Array1<T>> = self
+            .constellation
+            .get_node_input_bounds(current_node)
+            .map(|bounds| {
+                let fixed_bounds: Bounds1<T> = bounds.split_at(bounds.ndim() - unsafe_len).0;
+                let fixed_array: Array1<T> = fixed_bounds.lower().to_owned();
+                fixed_array
+            });
+        let mut rng = rand::thread_rng();
+        let mut best_sample = None;
         let mut best_val = T::infinity();
         let sample_opt: Option<(Array1<T>, T)> = if let Some(fixed_input_part) = fixed_input_part {
             let mut input_iter = unsafe_sample
                 .into_iter()
+                .filter(|x| match &input_set {
+                    Some(poly) => poly.is_member(&x.view()),
+                    None => true,
+                })
+                .filter(|x| {
+                    /*
+                    diag_gaussian_accept_reject(
+                        &x.mapv(|x| x.into()).view(),
+                        &self.constellation.get_loc().mapv(|x| x.into()).view(),
+                        &self
+                            .constellation
+                            .get_scale()
+                            .diag()
+                            .mapv(|x| x.into())
+                            .view(),
+                        &mut rng,
+                    )
+                    */
+                    let logp = gaussian_logp(
+                        &x.mapv(|x| x.into()).view(),
+                        &self.constellation.get_loc().mapv(|x| x.into()).view(),
+                        &self
+                            .constellation
+                            .get_scale()
+                            .diag()
+                            .mapv(|x| x.into())
+                            .view(),
+                    );
+                    logp > -1.8
+                })
                 .zip(iter::repeat(fixed_input_part))
-                .map(|(unfix, fix)| concatenate(Axis(0), &[unfix.view(), fix.view()]).unwrap());
+                .map(|(unfix, fix)| concatenate(Axis(0), &[fix.view(), unfix.view()]).unwrap());
             input_iter.find_map(|x| {
                 let output = self.constellation.get_dnn().forward1(x.clone());
                 debug_assert_eq!(output.len(), 1);
@@ -276,8 +328,8 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
                     Some((x, output[[0]]))
                 } else {
                     if output[[0]] < best_val {
-                        best_sample = x;
-                        best_val = output[[0]]
+                        best_sample = Some(x);
+                        best_val = output[[0]];
                     }
                     None
                 }
@@ -291,8 +343,8 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
                     Some((x, output[[0]]))
                 } else {
                     if output[[0]] < best_val {
-                        best_sample = x;
-                        best_val = output[[0]]
+                        best_sample = Some(x);
+                        best_val = output[[0]];
                     }
                     None
                 }
@@ -300,17 +352,19 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
         };
         sample_opt
             .map(|(sample, est_cost)| {
-                let unfixed = sample.slice(s![..unsafe_len]).to_owned();
+                let unfixed = sample.slice(s![-(unsafe_len as isize)..]).to_owned();
                 (unfixed, est_cost)
             })
-            .ok_or_else(|| (best_sample.slice(s![..unsafe_len]).to_owned(), best_val))
+            .ok_or_else(|| {
+                best_sample.map(|x| (x.slice(s![-(unsafe_len as isize)..]).to_owned(), best_val))
+            })
     }
 
     /// Given a node, samples one of its children and returns the log probability
     ///
     /// Returns:
-    ///     child_idx (usize): The child that was reached in the sampling procedure
-    ///     path_logp (T): The log probability of reaching the child from the root node
+    ///     `child_idx` (usize): The child that was reached in the sampling procedure
+    ///     `path_logp` (T): The log probability of reaching the child from the root node
     ///
     /// # Panics
     fn select_safe_child<R: Rng>(
@@ -320,13 +374,15 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
         cdf_samples: usize,
         max_iters: usize,
         rng: &mut R,
+        stability_eps: T,
     ) -> Option<(usize, T)> {
-        let input_bounds = self.constellation.get_input_bounds();
         debug_assert!(self
             .constellation
             .try_get_gaussian_distribution(current_node)
             .is_some());
-        let current_node_type = self.constellation.get_node_type(current_node);
+        let current_node_type = self
+            .constellation
+            .get_node_type(current_node, stability_eps);
         match *current_node_type {
             // leaf node, which must be partially safe and partially unsafe
             StarNodeType::Leaf => {
@@ -334,9 +390,9 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
             }
             StarNodeType::Affine { child_idx } => Some((child_idx, path_logp)),
             StarNodeType::StepRelu {
-                dim,
                 fst_child_idx,
                 snd_child_idx,
+                ..
             } => {
                 let mut children = vec![fst_child_idx];
                 if let Some(snd_child) = snd_child_idx {
@@ -353,15 +409,17 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
                     2 => {
                         let fst_cdf = self.constellation.get_node_cdf(
                             children[0],
-                            cdf_samples,
+                            cdf_samples * 4,
                             max_iters,
                             rng,
+                            stability_eps,
                         );
                         let snd_cdf = self.constellation.get_node_cdf(
                             children[1],
                             cdf_samples,
                             max_iters,
                             rng,
+                            stability_eps,
                         );
                         debug!(
                             "Selecting between 2 children with CDFs: {} and {}",
@@ -383,6 +441,7 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
                                     cdf_samples,
                                     max_iters,
                                     rng,
+                                    stability_eps,
                                 );
                                 let mut derived_fst_cdf = parent_cdf - snd_cdf;
 
@@ -402,6 +461,7 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
                                     cdf_samples,
                                     max_iters,
                                     rng,
+                                    stability_eps,
                                 );
                                 let mut derived_snd_cdf = parent_cdf - fst_cdf;
                                 if derived_snd_cdf.is_sign_negative() {
@@ -432,11 +492,11 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
                 }
             }
             StarNodeType::StepReluDropOut {
-                dim,
                 dropout_prob,
                 fst_child_idx,
                 snd_child_idx,
                 trd_child_idx,
+                ..
             } => {
                 let dropout_dist = Bernoulli::new(dropout_prob.into()).unwrap();
                 if dropout_dist.sample(rng) {
@@ -466,12 +526,14 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
                                 cdf_samples,
                                 max_iters,
                                 rng,
+                                stability_eps,
                             );
                             let snd_cdf = self.constellation.get_node_cdf(
                                 children[1],
                                 cdf_samples,
                                 max_iters,
                                 rng,
+                                stability_eps,
                             );
 
                             // Handle the case where a CDF gives a non-normal value
@@ -483,6 +545,7 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
                                         cdf_samples,
                                         max_iters,
                                         rng,
+                                        stability_eps,
                                     );
                                     let derived_fst_cdf = parent_cdf - snd_cdf;
                                     self.constellation
@@ -495,6 +558,7 @@ impl<'a, T: NNVFloat> Asterism<'a, T, Ix2> {
                                         cdf_samples,
                                         max_iters,
                                         rng,
+                                        stability_eps,
                                     );
                                     let derived_snd_cdf = parent_cdf - fst_cdf;
                                     self.constellation
@@ -535,7 +599,7 @@ mod test {
             let mut rng = rand::thread_rng();
             let mut asterism = Asterism::new(&mut constellation, 1.);
             let default: Array1<f64> = Array1::zeros(asterism.constellation.get_dnn().input_shape()[0].unwrap());
-            let sample = asterism.sample_safe_star(1, &mut rng, 1, 1, None);
+            let sample = asterism.sample_safe_star(1, &mut rng, 1, 1, None, 1e-4);
         }
     }
 }
