@@ -1,14 +1,16 @@
 #![allow(non_snake_case)]
-use crate::bounds::Bounds;
+use crate::bounds::Bounds1;
 use crate::dnn::DNNIterator;
 use crate::dnn::{DNNIndex, DNN};
-use crate::polytope::PolytopeInputDistribution;
+use crate::gaussian::GaussianDistribution;
+use crate::polytope::Polytope;
 use crate::star::Star;
 use crate::star_node::StarNode;
 use crate::star_node::StarNodeOp;
 use crate::star_node::StarNodeType;
 use crate::util::ArenaLike;
 use crate::NNVFloat;
+use ndarray::ArrayView1;
 use ndarray::Dimension;
 use ndarray::Ix2;
 use ndarray::{Array1, Array2};
@@ -23,18 +25,11 @@ pub struct Constellation<T: NNVFloat, D: Dimension> {
     loc: Array1<T>,
     scale: Array2<T>,
     dnn: DNN<T>,
-    input_bounds: Option<Bounds<T, D>>,
 }
 
 impl<T: NNVFloat, D: Dimension> Constellation<T, D> {
     /// Instantiate a Constellation with given input set and network
-    pub fn new(
-        input_star: Star<T, D>,
-        dnn: DNN<T>,
-        input_bounds: Option<Bounds<T, D>>,
-        loc: Array1<T>,
-        scale: Array2<T>,
-    ) -> Self {
+    pub fn new(input_star: Star<T, D>, dnn: DNN<T>, loc: Array1<T>, scale: Array2<T>) -> Self {
         let star_node = StarNode::default(input_star, None);
         let node_type = vec![None];
         let arena = vec![star_node];
@@ -46,7 +41,6 @@ impl<T: NNVFloat, D: Dimension> Constellation<T, D> {
             loc,
             scale,
             dnn,
-            input_bounds,
         }
     }
 
@@ -66,75 +60,78 @@ impl<T: NNVFloat, D: Dimension> Constellation<T, D> {
         &self.scale
     }
 
-    pub fn get_input_bounds(&self) -> &Option<Bounds<T, D>> {
-        &self.input_bounds
-    }
-
-    /*
-    fn add_node(&mut self, node: StarNode<T, D>, parent_id: usize) -> usize {
-        let child_idx = self.arena.new_node(node);
-        let other_child_idx = self.node_type.new_node(None);
-        let other_other_child_idx = self.parents.new_node(Some(parent_id));
-        debug_assert_eq!(child_idx, other_child_idx);
-        child_idx
-    }
-    */
-
     pub fn reset_input_distribution(&mut self, loc: Array1<T>, scale: Array2<T>) {
         self.loc = loc;
         self.scale = scale;
-        self.arena.iter_mut().for_each(|x| x.reset_cdf());
+        self.arena.iter_mut().for_each(StarNode::reset_cdf);
     }
 
-    pub fn reset_with_star(&mut self, input_star: Star<T, D>, input_bounds: Option<Bounds<T, D>>) {
+    pub fn reset_with_star(&mut self, input_star: Star<T, D>) {
         let star_node = StarNode::default(input_star, None);
         self.arena = vec![star_node];
         self.node_type = vec![None];
-        self.input_bounds = input_bounds;
+        self.parents = vec![None];
     }
 }
 
 impl<T: crate::NNVFloat> Constellation<T, Ix2> {
-    fn add_node_with_gauss_init(&mut self, node: StarNode<T, Ix2>, parent_id: usize) -> usize {
+    pub fn get_input_bounds(&self) -> Option<&Bounds1<T>> {
+        self.arena[0].get_input_bounds()
+    }
+
+    fn add_node_with_gauss_init(
+        &mut self,
+        node: StarNode<T, Ix2>,
+        parent_id: usize,
+        stability_eps: T,
+    ) -> usize {
         let child_idx = self.arena.new_node(node);
         let other_child_idx = self.node_type.new_node(None);
         let other_other_child_idx = self.parents.new_node(Some(parent_id));
         debug_assert_eq!(child_idx, other_child_idx);
+        debug_assert_eq!(child_idx, other_other_child_idx);
         debug_assert!(self.try_get_gaussian_distribution(parent_id).is_some());
-        self.initialize_node_tilting_from_parent(child_idx, parent_id, 20);
+        self.initialize_node_tilting_from_parent(child_idx, parent_id, 20, stability_eps);
         child_idx
     }
 
+    /// # Panics
     pub fn initialize_node_tilting_from_parent(
         &mut self,
         node_id: usize,
         parent_id: usize,
         max_accept_reject_iters: usize,
+        stability_eps: T,
     ) {
         let parent_tilting_soln = {
             debug_assert!(self.try_get_gaussian_distribution(parent_id).is_some());
             let parent_distr = self.arena[parent_id]
                 .try_get_gaussian_distribution()
                 .unwrap();
-            debug_assert!(parent_distr.try_get_tilting_solution().is_some());
-            parent_distr.try_get_tilting_solution().unwrap().clone()
+            // This isn't always the case because the parent may be unconstrained (if it's the root)
+            //debug_assert!(parent_distr.try_get_tilting_solution().is_some());
+            parent_distr.try_get_tilting_solution().cloned()
         };
         let child_distr = self.arena[node_id].get_gaussian_distribution(
             &self.loc,
             &self.scale,
             max_accept_reject_iters,
-            &self.input_bounds,
+            stability_eps,
         );
-        child_distr.get_tilting_solution(Some(&parent_tilting_soln));
+        child_distr.populate_tilting_solution(parent_tilting_soln.as_ref());
     }
 
     pub fn try_get_gaussian_distribution(
         &self,
         node_id: usize,
-    ) -> Option<&PolytopeInputDistribution<T>> {
+    ) -> Option<&GaussianDistribution<T>> {
         self.arena[node_id].try_get_gaussian_distribution()
     }
 
+    pub fn get_node_input_bounds(&self, node_id: usize) -> Option<&Bounds1<T>> {
+        self.arena[node_id].get_input_bounds()
+    }
+
     pub fn get_node_dnn_index(&self, node_id: usize) -> DNNIndex {
         self.arena[node_id].get_dnn_index()
     }
@@ -143,8 +140,16 @@ impl<T: crate::NNVFloat> Constellation<T, Ix2> {
         self.arena[node_id].get_output_bounds(&self.dnn, &|x| (x.lower()[[0]], x.upper()[[0]]))
     }
 
-    pub fn is_node_leaf(&mut self, id: usize) -> bool {
-        *self.get_node_type(id) == StarNodeType::Leaf
+    pub fn is_node_leaf(&mut self, id: usize, stability_eps: T) -> bool {
+        *self.get_node_type(id, stability_eps) == StarNodeType::Leaf
+    }
+
+    pub fn get_node_reduced_input_polytope(&self, node_id: usize) -> Option<Polytope<T>> {
+        self.arena[node_id].get_reduced_input_polytope()
+    }
+
+    pub fn is_node_member(&self, node_id: usize, point: &ArrayView1<T>) -> bool {
+        self.arena[node_id].is_input_member(point)
     }
 
     pub fn get_node_cdf<R: Rng>(
@@ -153,14 +158,15 @@ impl<T: crate::NNVFloat> Constellation<T, Ix2> {
         cdf_samples: usize,
         max_iters: usize,
         rng: &mut R,
+        stability_eps: T,
     ) -> T {
         self.arena[node_id].gaussian_cdf(
             &self.loc,
             &self.scale,
             cdf_samples,
             max_iters,
-            &self.input_bounds,
             rng,
+            stability_eps,
         )
     }
 
@@ -178,6 +184,7 @@ impl<T: crate::NNVFloat> Constellation<T, Ix2> {
         rng: &mut R,
         n: usize,
         max_iters: usize,
+        stability_eps: T,
     ) -> Vec<Array1<T>> {
         let node = &mut self.arena[node_id];
         node.gaussian_sample(
@@ -186,28 +193,29 @@ impl<T: crate::NNVFloat> Constellation<T, Ix2> {
             &self.scale,
             n,
             max_iters,
-            &self.input_bounds,
-            None,
+            &None,
+            stability_eps,
         )
         .into_iter()
         .map(|input| node.forward(&input))
         .collect()
     }
 
+    /// # Panics
     pub fn sample_gaussian_node<R: Rng>(
         &mut self,
         node_id: usize,
         rng: &mut R,
         n: usize,
         max_iters: usize,
+        stability_eps: T,
     ) -> Vec<Array1<T>> {
-        let initialization_opt = self.parents[node_id].map(|x| {
+        let initialization_opt = self.parents[node_id].and_then(|x| {
             self.arena[x]
                 .try_get_gaussian_distribution()
                 .unwrap()
                 .try_get_tilting_solution()
-                .unwrap()
-                .clone()
+                .cloned()
         });
         self.arena[node_id].gaussian_sample(
             rng,
@@ -215,11 +223,12 @@ impl<T: crate::NNVFloat> Constellation<T, Ix2> {
             &self.scale,
             n,
             max_iters,
-            &self.input_bounds,
-            initialization_opt,
+            &initialization_opt,
+            stability_eps,
         )
     }
 
+    /// # Panics
     pub fn sample_gaussian_node_safe<R: Rng>(
         &mut self,
         node_id: usize,
@@ -227,6 +236,7 @@ impl<T: crate::NNVFloat> Constellation<T, Ix2> {
         n: usize,
         max_iters: usize,
         safe_value: T,
+        stability_eps: T,
     ) -> Vec<Array1<T>> {
         let mut safe_star = self.arena[node_id].get_safe_star(safe_value);
         let initialization_opt = self.parents[node_id].map(|x| {
@@ -243,8 +253,8 @@ impl<T: crate::NNVFloat> Constellation<T, Ix2> {
             &self.scale,
             n,
             max_iters,
-            &self.input_bounds,
-            initialization_opt,
+            &initialization_opt,
+            stability_eps,
         )
     }
 
@@ -262,7 +272,7 @@ impl<T: crate::NNVFloat> Constellation<T, Ix2> {
     /// * `children` - `StarNodeType<T>`
     ///
     /// # Panics
-    pub fn get_node_type(&mut self, node_id: usize) -> &StarNodeType<T> {
+    pub fn get_node_type(&mut self, node_id: usize, stability_eps: T) -> &StarNodeType<T> {
         if self
             .node_type
             .get(node_id)
@@ -275,12 +285,12 @@ impl<T: crate::NNVFloat> Constellation<T, Ix2> {
                 .unwrap()
         } else {
             debug_assert!(self.try_get_gaussian_distribution(node_id).is_some());
-            self.expand(node_id)
+            self.expand(node_id, stability_eps)
         }
     }
 
-    pub fn get_node_child_ids(&mut self, node_id: usize) -> Vec<usize> {
-        match self.get_node_type(node_id) {
+    pub fn get_node_child_ids(&mut self, node_id: usize, stability_eps: T) -> Vec<usize> {
+        match self.get_node_type(node_id, stability_eps) {
             StarNodeType::Leaf => vec![],
             StarNodeType::Affine { child_idx } => vec![*child_idx],
             StarNodeType::StepRelu {
@@ -295,11 +305,10 @@ impl<T: crate::NNVFloat> Constellation<T, Ix2> {
                 child_ids
             }
             StarNodeType::StepReluDropOut {
-                dim,
-                dropout_prob,
                 fst_child_idx,
                 snd_child_idx,
                 trd_child_idx,
+                ..
             } => {
                 let mut child_ids: Vec<usize> = vec![*fst_child_idx];
                 if let Some(idx) = snd_child_idx {
@@ -323,7 +332,7 @@ impl<T: crate::NNVFloat> Constellation<T, Ix2> {
     ///
     /// # Returns
     /// * `children` - `StarNodeType<T>`
-    fn expand(&mut self, node_id: usize) -> &StarNodeType<T> {
+    fn expand(&mut self, node_id: usize, stability_eps: T) -> &StarNodeType<T> {
         let dnn_index = self.arena[node_id].get_dnn_index();
         let dnn_iter = &mut DNNIterator::new(&self.dnn, dnn_index);
 
@@ -340,6 +349,7 @@ impl<T: crate::NNVFloat> Constellation<T, Ix2> {
                     StarNode::default(self.arena[node_id].get_star().affine_map2(&aff), None)
                         .with_dnn_index(dnn_iter.get_idx()),
                     node_id,
+                    stability_eps,
                 );
                 StarNodeType::Affine { child_idx }
             }
@@ -350,24 +360,24 @@ impl<T: crate::NNVFloat> Constellation<T, Ix2> {
                 let mut ids = vec![];
 
                 if let Some(lower_star) = child_stars.0 {
-                    let mut bounds = self.arena[node_id].get_local_bounds().clone();
+                    let mut bounds = self.arena[node_id].calculate_star_local_bounds().clone();
                     bounds.index_mut(dim)[0] = T::zero();
                     bounds.index_mut(dim)[1] = T::zero();
                     let lower_node =
                         StarNode::default(lower_star, Some(bounds)).with_dnn_index(dnn_idx);
-                    let id = self.add_node_with_gauss_init(lower_node, node_id);
+                    let id = self.add_node_with_gauss_init(lower_node, node_id, stability_eps);
                     ids.push(id);
                 }
 
                 if let Some(upper_star) = child_stars.1 {
-                    let mut bounds = self.arena[node_id].get_local_bounds().clone();
+                    let mut bounds = self.arena[node_id].calculate_star_local_bounds().clone();
                     let mut lb = bounds.index_mut(dim);
                     if lb[0].is_sign_negative() {
                         lb[0] = T::zero();
                     }
                     let upper_node =
                         StarNode::default(upper_star, Some(bounds)).with_dnn_index(dnn_idx);
-                    let id = self.add_node_with_gauss_init(upper_node, node_id);
+                    let id = self.add_node_with_gauss_init(upper_node, node_id, stability_eps);
                     ids.push(id);
                 }
 
@@ -383,34 +393,34 @@ impl<T: crate::NNVFloat> Constellation<T, Ix2> {
                 let mut ids = vec![];
 
                 if let Some(dropout_star) = child_stars.0 {
-                    let mut bounds = self.arena[node_id].get_local_bounds().clone();
+                    let mut bounds = self.arena[node_id].calculate_star_local_bounds().clone();
                     bounds.index_mut(dim)[0] = T::zero();
                     bounds.index_mut(dim)[1] = T::zero();
                     let dropout_node =
                         StarNode::default(dropout_star, Some(bounds)).with_dnn_index(dnn_idx);
-                    let id = self.add_node_with_gauss_init(dropout_node, node_id);
+                    let id = self.add_node_with_gauss_init(dropout_node, node_id, stability_eps);
                     ids.push(id);
                 }
 
                 if let Some(lower_star) = child_stars.1 {
-                    let mut bounds = self.arena[node_id].get_local_bounds().clone();
+                    let mut bounds = self.arena[node_id].calculate_star_local_bounds().clone();
                     bounds.index_mut(dim)[0] = T::zero();
                     bounds.index_mut(dim)[1] = T::zero();
                     let lower_node =
                         StarNode::default(lower_star, Some(bounds)).with_dnn_index(dnn_idx);
-                    let id = self.add_node_with_gauss_init(lower_node, node_id);
+                    let id = self.add_node_with_gauss_init(lower_node, node_id, stability_eps);
                     ids.push(id);
                 }
 
                 if let Some(upper_star) = child_stars.2 {
-                    let mut bounds = self.arena[node_id].get_local_bounds().clone();
+                    let mut bounds = self.arena[node_id].calculate_star_local_bounds().clone();
                     let mut lb = bounds.index_mut(dim);
                     if lb[0].is_sign_negative() {
                         lb[0] = T::zero();
                     }
                     let upper_node =
                         StarNode::default(upper_star, Some(bounds)).with_dnn_index(dnn_idx);
-                    let id = self.add_node_with_gauss_init(upper_node, node_id);
+                    let id = self.add_node_with_gauss_init(upper_node, node_id, stability_eps);
                     ids.push(id);
                 }
 
@@ -443,13 +453,13 @@ mod test {
         fn test_cache_bounds_equivalence(mut constellation in generic_constellation(2, 2, 2, 2)) {
             let mut rng = rand::thread_rng();
             let mut asterism = Asterism::new(&mut constellation, 1.);
-            asterism.sample_safe_star(1, &mut rng, 1, 1, None);
+            asterism.sample_safe_star(1, &mut rng, 1, 1, None, 1e-4);
 
             for (i, node) in constellation.arena.iter_mut().enumerate() {
-                if let Some(local_bounds) = node.get_local_bounds_direct() {
+                if let Some(local_bounds) = node.calculate_star_local_bounds_direct() {
                     let bounds = local_bounds.clone();
                     node.set_local_bounds_direct(None);
-                    let expected_local_bounds = node.get_local_bounds();
+                    let expected_local_bounds = node.calculate_star_local_bounds();
 
                     prop_assert!(bounds.bounds_iter().into_iter()
                         .zip(expected_local_bounds.bounds_iter().into_iter())
