@@ -1,6 +1,7 @@
 use crate::affine::Affine2;
-use crate::util::solve;
-use crate::util::LinearSolution;
+use crate::bounds::Bounds1;
+use crate::lp::solve;
+use crate::lp::LinearSolution;
 use crate::NNVFloat;
 use ndarray::arr1;
 use ndarray::concatenate;
@@ -18,12 +19,17 @@ use std::ops::MulAssign;
 pub struct Inequality<T: NNVFloat> {
     coeffs: Array2<T>, // Assume rows correspond to equations and cols are vars, i.e. Ax < b
     rhs: Array1<T>,
+    bounds: Bounds1<T>,
 }
 
 impl<T: NNVFloat> Inequality<T> {
-    pub fn new(coeffs: Array2<T>, rhs: Array1<T>) -> Self {
+    pub fn new(coeffs: Array2<T>, rhs: Array1<T>, bounds: Bounds1<T>) -> Self {
         debug_assert_eq!(coeffs.nrows(), rhs.len());
-        Self { coeffs, rhs }
+        Self {
+            coeffs,
+            rhs,
+            bounds,
+        }
     }
 
     pub fn coeffs(&self) -> ArrayView2<T> {
@@ -46,6 +52,10 @@ impl<T: NNVFloat> Inequality<T> {
         self.rhs.len()
     }
 
+    pub fn bounds(&self) -> &Bounds1<T> {
+        &self.bounds
+    }
+
     pub fn check_redundant(&self, coeffs: ArrayView1<T>, rhs: T) -> bool {
         let neg_one: T = std::convert::From::from(-1.);
         let maximize_eqn = &coeffs * neg_one;
@@ -57,18 +67,19 @@ impl<T: NNVFloat> Inequality<T> {
                 .chain(iter::once(maximize_eqn.view())),
             concatenate![Axis(0), self.rhs(), arr1(&[maximize_rhs])].view(),
             maximize_eqn.view(),
+            &self.bounds,
         );
         let val: f64 = match solved {
             LinearSolution::Solution(_, val) => val,
-            LinearSolution::Infeasible => return true,
-            LinearSolution::Unbounded => return true,
+            LinearSolution::Infeasible | LinearSolution::Unbounded(_) => return true,
         };
         val > rhs.into()
     }
 
+    /// `check_redundant` is currently disabled
     /// # Panics
     pub fn add_eqns(&mut self, eqns: &Self, check_redundant: bool) {
-        if check_redundant {
+        if check_redundant && false {
             for (c, r) in eqns.coeffs().rows().into_iter().zip(eqns.rhs().into_iter()) {
                 if !self.check_redundant(c, *r) {
                     self.coeffs
@@ -79,9 +90,10 @@ impl<T: NNVFloat> Inequality<T> {
                         .expect("Failed to add eqn to Inequality");
                 }
             }
+        } else {
+            self.coeffs.append(Axis(0), eqns.coeffs()).unwrap();
+            self.rhs.append(Axis(0), eqns.rhs()).unwrap();
         }
-        self.coeffs.append(Axis(0), eqns.coeffs()).unwrap();
-        self.rhs.append(Axis(0), eqns.rhs()).unwrap();
     }
 
     pub fn any_nan(&self) -> bool {
@@ -113,9 +125,11 @@ impl<T: NNVFloat> Inequality<T> {
                 .rhs
                 .slice_axis(Axis(0), Slice::new(i_idx, Some(i_idx + 1), 1))
                 .to_owned(),
+            bounds: self.bounds.get_ith_bounds(idx),
         }
     }
 
+    /// Returns whether a given point is in the set represented by the `Inequality`
     pub fn is_member(&self, point: &ArrayView1<T>) -> bool {
         let vals = self.coeffs.dot(point);
         Zip::from(&self.rhs)
@@ -123,47 +137,97 @@ impl<T: NNVFloat> Inequality<T> {
             .fold(true, |acc, ub, v| acc && (v <= ub))
     }
 
-    /// Assumes that the zero valued
+    /// Remove dimensions from the inequality that have fixed value.
     ///
-    /// TODO: Test this function
+    /// # Arguments
+    ///
+    /// * `x` - An Array with fixed values in each dimension. Any dimensions that aren't fixed should be set to zero.
+    /// * `fixed_idxs` - Array that indicates which dimensions are fixed with `true` (because a dim could be fixed at zero)
     ///
     /// # Returns
-    /// None if all the indices are removed
+    ///
+    /// None if all the indices are fixed, otherwise a `Self` with reduced dimension
     ///
     /// # Panics
+    ///
+    /// Only if the underlying struct is malformed
     pub fn reduce_with_values(
         &self,
         x: ArrayView1<T>,
         fixed_idxs: ArrayView1<bool>,
     ) -> Option<Self> {
+        // Check if every dimension is fixed
         if fixed_idxs.iter().all(|y| *y) {
             return None;
         }
-        // Reduce the rhs of each constraint
-        let lhs_values: Array1<T> = self.coeffs.dot(&x);
-        let reduced_rhs = &self.rhs - lhs_values;
+
+        // Reduce the rhs of each constraint (but don't filter cuz no rows are removed)
+        let reduced_rhs: Array1<T> = &self.rhs - self.coeffs.dot(&x);
 
         // Remove the variables that are fixed
-        let filtered_coeffs: Vec<ArrayView2<T>> = self
-            .coeffs
-            .columns()
+        let filtered_coeffs = {
+            // Specifically, remove columns
+            let filtered_cols: Vec<ArrayView2<T>> = self
+                .coeffs
+                .columns()
+                .into_iter()
+                .zip(fixed_idxs.iter())
+                .filter(|(_item, &is_fixed)| !is_fixed)
+                .map(|x| x.0.insert_axis(Axis(1)))
+                .collect();
+            // We know this unwrap won't error because we haven't changed the number of rows for any column
+            concatenate(Axis(1), &filtered_cols).unwrap()
+        };
+
+        let filtered_bounds_iter = self
+            .bounds()
+            .bounds_iter()
             .into_iter()
             .zip(&fixed_idxs)
-            .filter(|x| !*x.1)
-            .map(|x| x.0.insert_axis(Axis(1)))
+            .filter(|(_, &is_fixed)| !is_fixed)
+            .map(|(dim, _drop)| dim);
+
+        // Remove trivial constraints (i.e. all zero coeffs on LHS)
+        let is_nontrivial: Vec<bool> = filtered_coeffs
+            .rows()
+            .clone()
+            .into_iter()
+            .map(|row| row.iter().all(|x| !x.is_zero()))
             .collect();
-        let new_eqns = concatenate(Axis(1), filtered_coeffs.as_slice()).unwrap();
-        let (filtered_eqns, filtered_rhs): (Vec<ArrayView2<T>>, Vec<T>) = new_eqns
+
+        // Handle the case where every constraint is trivial
+        if !is_nontrivial.iter().any(|y| *y) {
+            return None;
+        }
+
+        let nontrivial_coeffs: Vec<ArrayView2<T>> = filtered_coeffs
             .rows()
             .into_iter()
-            .zip(reduced_rhs.into_iter())
-            .filter(|(coeffs, _rhs)| coeffs.iter().any(|x| !x.is_zero()))
-            .map(|(coeffs, rhs)| (coeffs.insert_axis(Axis(0)), rhs))
-            .unzip();
-        let newer_eqns: Array2<T> = concatenate(Axis(0), filtered_eqns.as_slice()).unwrap();
-        let filtered_rhs = Array1::from_vec(filtered_rhs);
+            .zip(is_nontrivial.iter())
+            .filter(|(_, &is_nontrivial)| is_nontrivial)
+            .map(|(row, _)| row.insert_axis(Axis(0)))
+            .collect();
+        let nontrivial_rhs: Vec<T> = reduced_rhs
+            .into_iter()
+            .zip(is_nontrivial.iter())
+            .filter(|(_, &is_nontrivial)| is_nontrivial)
+            .map(|(val, _)| val)
+            .collect();
+        let nontrivial_bounds: Vec<ArrayView1<T>> = filtered_bounds_iter
+            .zip(is_nontrivial.iter())
+            .filter(|(_, &is_nontrivial)| is_nontrivial)
+            .map(|(val, _)| val)
+            .collect();
 
-        Some(Self::new(newer_eqns, filtered_rhs))
+        let final_coeffs: Array2<T> = concatenate(Axis(0), &nontrivial_coeffs).unwrap();
+        let final_rhs = Array1::from_vec(nontrivial_rhs);
+        let final_bounds = if nontrivial_bounds.is_empty() {
+            Bounds1::trivial(0)
+        } else {
+            Bounds1::new_by_dim(nontrivial_bounds.as_slice())
+        };
+
+        Some(Self::new(final_coeffs, final_rhs, final_bounds))
     }
 }
 
@@ -173,8 +237,9 @@ impl<T: NNVFloat> Mul<T> for Inequality<T> {
 
     fn mul(self, rhs: T) -> Self {
         Self {
-            coeffs: &self.coeffs * rhs,
-            rhs: &self.rhs * rhs,
+            coeffs: self.coeffs * rhs,
+            rhs: self.rhs * rhs,
+            bounds: self.bounds * rhs,
         }
     }
 }
@@ -184,6 +249,7 @@ impl<T: NNVFloat> MulAssign<T> for Inequality<T> {
     fn mul_assign(&mut self, rhs: T) {
         self.coeffs *= rhs;
         self.rhs *= rhs;
+        self.bounds *= rhs;
     }
 }
 
@@ -192,6 +258,7 @@ impl<T: NNVFloat> From<Affine2<T>> for Inequality<T> {
         Self {
             coeffs: aff.basis().to_owned(),
             rhs: aff.shift().to_owned(),
+            bounds: Bounds1::trivial(aff.shift().len()),
         }
     }
 }
