@@ -1,28 +1,19 @@
 #![allow(non_snake_case, clippy::similar_names)]
 //! Implementation of H-representation polytopes
 use crate::bounds::Bounds1;
+use crate::gaussian::GaussianDistribution;
 use crate::inequality::Inequality;
+use crate::lp::solve;
+use crate::lp::LinearSolution;
 use crate::util;
-use crate::util::l2_norm;
-use crate::util::solve;
-use crate::util::LinearExpression;
-use crate::util::LinearSolution;
 use crate::NNVFloat;
-use good_lp::solvers::coin_cbc::coin_cbc;
-use good_lp::Expression;
-use good_lp::ProblemVariables;
-use good_lp::Variable;
-use good_lp::{variable, Solution, SolverModel};
-use ndarray::concatenate;
+use ndarray::Array1;
 use ndarray::Array2;
 use ndarray::ArrayView1;
 use ndarray::ArrayView2;
 use ndarray::Ix2;
-use ndarray::{Array1, Axis};
-use rand::Rng;
 use std::fmt::Debug;
 use truncnorm::distributions::MultivariateTruncatedNormal;
-use truncnorm::tilting::TiltingSolution;
 
 /// H-representation polytope
 #[derive(Clone, Debug)]
@@ -31,9 +22,9 @@ pub struct Polytope<T: NNVFloat> {
 }
 
 impl<T: NNVFloat> Polytope<T> {
-    pub fn new(constraint_coeffs: Array2<T>, upper_bounds: Array1<T>) -> Self {
+    pub fn new(constraint_coeffs: Array2<T>, upper_bounds: Array1<T>, bounds: Bounds1<T>) -> Self {
         Self {
-            halfspaces: Inequality::new(constraint_coeffs, upper_bounds),
+            halfspaces: Inequality::new(constraint_coeffs, upper_bounds, bounds),
         }
     }
 
@@ -60,8 +51,12 @@ impl<T: NNVFloat> Polytope<T> {
         self.halfspaces.rhs()
     }
 
+    pub fn get_bounds(&self) -> &Bounds1<T> {
+        self.halfspaces.bounds()
+    }
+
     pub fn add_constraints(&mut self, constraints: &Inequality<T>, check_redundant: bool) {
-        self.halfspaces.add_eqns(constraints, check_redundant)
+        self.halfspaces.add_eqns(constraints, check_redundant);
     }
 
     pub fn num_constraints(&self) -> usize {
@@ -73,7 +68,7 @@ impl<T: NNVFloat> Polytope<T> {
     }
 
     pub fn filter_trivial(&mut self) {
-        self.halfspaces.filter_trivial()
+        self.halfspaces.filter_trivial();
     }
 
     pub fn is_member(&self, point: &ArrayView1<T>) -> bool {
@@ -85,7 +80,8 @@ impl<T: NNVFloat> Polytope<T> {
         mu: &Array1<T>,
         sigma: &Array2<T>,
         max_accept_reject_iters: usize,
-    ) -> PolytopeInputDistribution<T> {
+        stability_eps: T,
+    ) -> GaussianDistribution<T> {
         // convert T to f64 in inputs
         let mu = mu.mapv(std::convert::Into::into);
         let sigma = sigma.mapv(std::convert::Into::into);
@@ -108,7 +104,8 @@ impl<T: NNVFloat> Polytope<T> {
 
         let sq_constr_sigma = {
             let sigma: Array2<f64> = constraint_coeffs.dot(&sigma.dot(&constraint_coeffs.t()));
-            let diag_addn = Array2::from_diag(&Array1::from_elem(sigma.nrows(), 1e-5));
+            let diag_addn: Array2<f64> =
+                Array2::from_diag(&Array1::from_elem(sigma.nrows(), stability_eps.into()));
             sigma + diag_addn
         };
         let sq_ub = ub;
@@ -122,14 +119,14 @@ impl<T: NNVFloat> Polytope<T> {
             sq_constr_ub,
             max_accept_reject_iters,
         );
-        let inv_coeffs: Array2<T> =
-            util::pinv(&constraint_coeffs.mapv(|x| x.into())).mapv(|x| x.into());
-        PolytopeInputDistribution {
+        let inv_coeffs: Array2<T> = util::pinv(&constraint_coeffs).mapv(Into::into);
+        GaussianDistribution::TruncGaussian {
             distribution,
             inv_coeffs,
         }
     }
 
+    /* Commented out while we're working on building the lp features
     /// Source: <https://stanford.edu/class/ee364a/lectures/problems.pdf>
     pub fn chebyshev_center(&self) -> Option<(Array1<f64>, f64)> {
         let b = self.halfspaces.rhs();
@@ -163,9 +160,11 @@ impl<T: NNVFloat> Polytope<T> {
             None
         }
     }
+    */
 
     /// Returns None if the reduced polytope is empty
-    pub fn reduce_fixed_inputs(&self, bounds: &Bounds1<T>) -> Option<Self> {
+    pub fn reduce_fixed_inputs(&self) -> Option<Self> {
+        let bounds = self.get_bounds();
         let fixed_idxs = bounds.fixed_idxs();
         let fixed_vals = bounds.fixed_vals_or_zeros();
 
@@ -194,6 +193,7 @@ impl<T: NNVFloat> Polytope<T> {
             self.halfspaces.coeffs().rows(),
             self.halfspaces.rhs(),
             c.view(),
+            self.get_bounds(),
         );
 
         !matches!(solved, LinearSolution::Solution(_, _))
@@ -204,56 +204,9 @@ impl<T: NNVFloat> Polytope<T> {
 #[allow(clippy::fallible_impl_from)]
 impl<T: NNVFloat> From<Bounds1<T>> for Polytope<T> {
     fn from(item: Bounds1<T>) -> Self {
-        let coeffs = concatenate(
-            Axis(0),
-            &[
-                Array2::eye(item.ndim()).view(),
-                (Array2::eye(item.ndim()) * T::neg(T::one())).view(),
-            ],
-        )
-        .unwrap();
-        let rhs = concatenate(
-            Axis(0),
-            &[
-                item.upper(),
-                (item.lower().to_owned() * T::neg(T::one())).view(),
-            ],
-        )
-        .unwrap();
-        let halfspaces = Inequality::new(coeffs, rhs);
+        let coeffs = Array2::zeros((1, item.ndim()));
+        let rhs = Array1::zeros(1);
+        let halfspaces = Inequality::new(coeffs, rhs, item);
         Self { halfspaces }
     }
 }
-
-#[derive(Debug, Clone)]
-pub struct PolytopeInputDistribution<T> {
-    distribution: MultivariateTruncatedNormal<Ix2>,
-    inv_coeffs: Array2<T>,
-}
-
-impl<T: NNVFloat> PolytopeInputDistribution<T> {
-    pub fn sample_n<R: Rng>(&mut self, n: usize, rng: &mut R) -> Vec<Array1<T>> {
-        let sample_arr = self.distribution.sample_n(n, rng);
-        sample_arr
-            .rows()
-            .into_iter()
-            .map(|x| (self.inv_coeffs.dot(&x.mapv(|x| x.into()))))
-            .collect()
-    }
-
-    pub fn cdf<R: Rng>(&mut self, n: usize, rng: &mut R) -> T {
-        let (est, _rel_err, _upper_bound) = self.distribution.cdf(n, rng);
-        est.into()
-    }
-
-    pub fn try_get_tilting_solution(&self) -> Option<&TiltingSolution> {
-        self.distribution.try_get_tilting_solution()
-    }
-
-    pub fn get_tilting_solution(
-        &mut self,
-        initialization: Option<&TiltingSolution>,
-    ) -> &TiltingSolution {
-        self.distribution.get_tilting_solution(initialization)
-    }
-}
