@@ -3,12 +3,12 @@
 //! for representing affine transformed sets
 use crate::affine::{Affine, Affine2, Affine4};
 use crate::bounds::Bounds1;
+use crate::gaussian::GaussianDistribution;
 use crate::inequality::Inequality;
+use crate::lp::solve;
+use crate::lp::LinearSolution;
 use crate::polytope::Polytope;
-use crate::polytope::PolytopeInputDistribution;
 use crate::tensorshape::TensorShape;
-use crate::util::solve;
-use crate::util::LinearSolution;
 use crate::NNVFloat;
 use ndarray::Array4;
 use ndarray::ArrayView1;
@@ -111,18 +111,14 @@ impl<T: NNVFloat> Star2<T> {
         mu: &Array1<T>,
         sigma: &Array2<T>,
         max_accept_reject_iters: usize,
-        input_bounds: &Option<Bounds1<T>>,
-    ) -> Option<PolytopeInputDistribution<T>> {
+        stability_eps: T,
+    ) -> Option<GaussianDistribution<T>> {
         self.constraints
             .as_ref()
-            .and_then(|x| {
-                if let Some(bounds) = input_bounds {
-                    x.reduce_fixed_inputs(bounds)
-                } else {
-                    Some(x.clone())
-                }
+            .and_then(Polytope::reduce_fixed_inputs)
+            .map(|poly| {
+                poly.get_truncnorm_distribution(mu, sigma, max_accept_reject_iters, stability_eps)
             })
-            .map(|poly| poly.get_truncnorm_distribution(mu, sigma, max_accept_reject_iters))
     }
 
     /// Create a new Star with given dimension.
@@ -169,6 +165,10 @@ impl<T: NNVFloat> Star2<T> {
 }
 
 impl<T: NNVFloat> Star2<T> {
+    pub fn get_input_bounds(&self) -> Option<&Bounds1<T>> {
+        self.constraints.as_ref().map(Polytope::get_bounds)
+    }
+
     pub fn with_input_bounds(mut self, input_bounds: Bounds1<T>) -> Self {
         if self.constraints.is_some() {
             self.constraints = self.constraints.map(|x| x.with_input_bounds(input_bounds));
@@ -269,20 +269,22 @@ impl<T: NNVFloat> Star2<T> {
     pub fn get_min(&self, idx: usize) -> T {
         let eqn = self.representation.get_eqn(idx);
 
-        if let Some(ref poly) = self.constraints {
-            let solved = solve(
-                poly.coeffs().rows(),
-                poly.ubs(),
-                eqn.basis().index_axis(Axis(0), 0),
-            );
-            if let LinearSolution::Solution(_, val) = solved {
-                eqn.shift()[0] + val.into()
-            } else {
-                panic!()
-            }
-        } else {
-            T::neg_infinity()
-        }
+        self.constraints.as_ref().map_or_else(
+            || T::neg_infinity(),
+            |poly| {
+                let solved = solve(
+                    poly.coeffs().rows(),
+                    poly.ubs(),
+                    eqn.basis().index_axis(Axis(0), 0),
+                    poly.get_bounds(),
+                );
+                if let LinearSolution::Solution(_, val) = solved {
+                    eqn.shift()[0] + val.into()
+                } else {
+                    panic!("Solution: {:?}", solved)
+                }
+            },
+        )
     }
 
     #[cfg(feature = "rayon")]
@@ -322,20 +324,22 @@ impl<T: NNVFloat> Star2<T> {
         let shift = eqn.shift()[0];
         eqn *= neg_one;
 
-        if let Some(ref poly) = self.constraints {
-            let solved = solve(
-                poly.coeffs().rows(),
-                poly.ubs(),
-                eqn.basis().index_axis(Axis(0), 0),
-            );
-            if let LinearSolution::Solution(_, val) = solved {
-                shift - val.into()
-            } else {
-                panic!()
-            }
-        } else {
-            T::infinity()
-        }
+        self.constraints.as_ref().map_or_else(
+            || T::infinity(),
+            |poly| {
+                let solved = solve(
+                    poly.coeffs().rows(),
+                    poly.ubs(),
+                    eqn.basis().index_axis(Axis(0), 0),
+                    poly.get_bounds(),
+                );
+                if let LinearSolution::Solution(_, val) = solved {
+                    shift - val.into()
+                } else {
+                    panic!()
+                }
+            },
+        )
     }
 
     pub fn calculate_axis_aligned_bounding_box(&self) -> Bounds1<T> {
@@ -350,91 +354,6 @@ impl<T: NNVFloat> Star2<T> {
             .as_ref()
             .map_or(false, crate::polytope::Polytope::is_empty)
     }
-
-    /*
-    /// # Returns
-    ///
-    /// (cdf estimate, estimate error, cdf upper bound)
-    ///
-    /// # Panics
-    pub fn trunc_gaussian_cdf<R: Rng>(
-        &self,
-        mu: &Array1<T>,
-        sigma: &Array2<T>,
-        n: usize,
-        max_iters: usize,
-        input_bounds: &Option<Bounds1<T>>,
-        rng: &mut R,
-    ) -> (f64, f64, f64) {
-        // remove fixed dimensions from mu and sigma
-        debug!(
-            "mu shape: {:?} sigma shape: {:?}",
-            mu.shape(),
-            sigma.shape()
-        );
-        if let Some(poly) = &self.constraints {
-            if let Some(bounds) = input_bounds {
-                let lbs = bounds.lower();
-                let ubs = bounds.upper();
-                debug_assert!(!poly.any_nan());
-                if let Some((mut reduced_poly, (_reduced_lbs, _reduced_ubs))) =
-                    poly.reduce_fixed_inputs(&lbs, &ubs)
-                {
-                    reduced_poly.filter_trivial();
-                    debug_assert!(!reduced_poly.any_nan());
-                    reduced_poly.gaussian_cdf(mu, sigma, n, max_iters, rng)
-                } else {
-                    (1., 0., 1.)
-                }
-            } else {
-                poly.gaussian_cdf(mu, sigma, n, max_iters, rng)
-            }
-        } else {
-            // Unbounded sums to 1
-            (1., 0., 1.)
-        }
-    }
-    */
-
-    /*
-    /// # Panics
-    ///
-    /// # Returns
-    /// Vec<(samples, pdf of sample)>
-    ///
-    ///
-    // Allow if_let_else because rng is used in both branches, so closures don't work
-    #[allow(clippy::too_many_lines, clippy::option_if_let_else)]
-    pub fn gaussian_sample<R: Rng>(
-        &self,
-        rng: &mut R,
-        mu: &Array1<T>,
-        sigma: &Array2<T>,
-        n: usize,
-        max_iters: usize,
-        input_bounds: &Option<Bounds1<T>>,
-    ) -> Vec<(Array1<f64>, f64)> {
-        // remove fixed dimensions from mu and sigma
-        if let Some(poly) = &self.constraints {
-            if let Some(bounds) = input_bounds {
-                let lbs = bounds.lower();
-                let ubs = bounds.upper();
-                if let Some((reduced_poly, (_reduced_lbs, _reduced_ubs))) =
-                    poly.reduce_fixed_inputs(&lbs, &ubs)
-                {
-                    reduced_poly.gaussian_sample(rng, mu, sigma, n, max_iters)
-                } else {
-                    vec![]
-                }
-            } else {
-                poly.gaussian_sample(rng, mu, sigma, n, max_iters)
-            }
-        } else {
-            // Unbounded sample from Gaussian
-            todo!()
-        }
-    }
-    */
 }
 
 impl<T: NNVFloat> Star4<T> {
@@ -463,138 +382,115 @@ impl<T: NNVFloat> Star4<T> {
 #[cfg(test)]
 mod test {
     use super::*;
-    use crate::test_util::{array1, array2, empty_star, non_empty_star, pos_def_array1};
-    use ndarray::{arr1, arr2};
+    use crate::test_util::{array2, empty_star, non_empty_star};
+    use ndarray::arr1;
     use proptest::prelude::*;
     use proptest::proptest;
-    use rand::prelude::*;
     use std::panic;
 
     /*
-    #[test]
-    fn test_gaussian_sample_manual() {
-        let mut rng = Pcg64::seed_from_u64(2);
-        let loc: Array1<f64> = arr1(&[-2.52]);
-        let scale_diag: Array1<f64> = arr1(&[0.00001]);
-        let lbs = loc.clone() - 3.5 * scale_diag.clone();
-        let ubs = loc.clone() + 3.5 * scale_diag.clone();
-        let input_bounds = Bounds1::new(lbs.view(), ubs.view());
-        let mut star: Star2<f64> = Star2::new(Array2::eye(1) * 8.016, Array1::zeros(1));
-        star = star.add_constraints(&Inequality::new(arr2(&[[-1.], [1.]]), arr1(&[2.52, 0.10])));
-        star.gaussian_sample(
-            &mut rng,
-            &loc,
-            &Array2::from_diag(&scale_diag).to_owned(),
-            10,
-            20,
-            &Some(input_bounds),
-        );
-        /*
-        Star { representation: Affine { basis: [[8.016197283509424]], shape=[1, 1], strides=[1, 1], layout=CFcf (0xf), const ndim=2, shift: [0.0], shape=[1], strides=[1], layout=CFcf (0xf), const ndim=1 }, constraints: Some(Polytope { halfspaces: Inequality { coeffs: [[-1.0],
-            [1.0],
-            [-1.0],
-            [1.0],
-            [-1.0]], shape=[5, 1], strides=[1, 1], layout=CFcf (0xf), const ndim=2, rhs: [2.52102610837301, 0.1001445177130087, 5.47728758574745, 20.0, 20.0], shape=[5], strides=[1], layout=CFcf (0xf), const ndim=1 } }) }
-        */
-    }
-    */
-
-    proptest! {
-        /*
         #[test]
-        fn test_gaussian_sample(star in non_empty_star(1,3), loc in array1(1), scale_diag in pos_def_array1(1)) {
-            let mut rng = rand::thread_rng();
+        fn test_gaussian_sample_manual() {
+            let mut rng = Pcg64::seed_from_u64(2);
+            let loc: Array1<f64> = arr1(&[-2.52]);
+            let scale_diag: Array1<f64> = arr1(&[0.00001]);
             let lbs = loc.clone() - 3.5 * scale_diag.clone();
             let ubs = loc.clone() + 3.5 * scale_diag.clone();
             let input_bounds = Bounds1::new(lbs.view(), ubs.view());
-            star.gaussian_sample(&mut rng, &loc, &Array2::from_diag(&scale_diag).to_owned(), 10, 20, &Some(input_bounds));
+            let mut star: Star2<f64> = Star2::new(Array2::eye(1) * 8.016, Array1::zeros(1));
+            star = star.add_constraints(&Inequality::new(arr2(&[[-1.], [1.]]), arr1(&[2.52, 0.10])));
+            star.gaussian_sample(
+                &mut rng,
+                &loc,
+                &Array2::from_diag(&scale_diag).to_owned(),
+                10,
+                20,
+                &Some(input_bounds),
+            );
+            /*
+            Star { representation: Affine { basis: [[8.016197283509424]], shape=[1, 1], strides=[1, 1], layout=CFcf (0xf), const ndim=2, shift: [0.0], shape=[1], strides=[1], layout=CFcf (0xf), const ndim=1 }, constraints: Some(Polytope { halfspaces: Inequality { coeffs: [[-1.0],
+                [1.0],
+                [-1.0],
+                [1.0],
+                [-1.0]], shape=[5, 1], strides=[1, 1], layn
+                let lbs = loc.clone() - 3.5 * scale_diag.clone();
+                let ubs = loc.clone() + 3.5 * scale_diag.clone();
+                let input_bounds = Bounds1::new(lbs.view(), ubs.view());
+                star.gaussian_sample(&mut rng, &loc, &Array2::from_diag(&scale_diag).to_owned(), 10, 20, &Some(input_bounds));
+            }
+            */
+
+            #[test]
+            fn test_get_min_feasible(star in non_empty_star(2,3)) {
+                prop_assert!(!star.input_space_polytope().unwrap().is_empty(), "Polytope is empty");
+                prop_assert!(!star.is_empty(), "Non empty star is empty");
+                let result = panic::catch_unwind(|| {
+                    star.get_min(0);
+                });
+                prop_assert!(result.is_ok(), "Calculating min resulted in panic for feasible star");
+            }
+
+            #[test]
+            fn test_get_min_infeasible(star in empty_star(2,1)) {
+                prop_assert!(star.input_space_polytope().unwrap().is_empty(), "Polytope is not empty");
+                prop_assert!(star.is_empty(), "Empty star is not empty");
+                let result = panic::catch_unwind(|| {
+                    star.get_min(0)
+                });
+                prop_assert!(result.is_err(), "Infeasible star did not panic for get_min {:#?}", result);
+            }
+
+            #[test]
+            fn test_get_max_feasible(star in non_empty_star(2,3)) {
+                prop_assert!(!star.input_space_polytope().unwrap().is_empty(), "Polytope is empty");
+                prop_assert!(!star.is_empty(), "Non empty star is empty");
+                let result = panic::catch_unwind(|| {
+                    star.get_max(0);
+                });
+                prop_assert!(result.is_ok(), "Calculating min resulted in panic for feasible star");
+            }
+
+            #[test]
+            fn test_get_max_infeasible(star in empty_star(2,1)) {
+                prop_assert!(star.input_space_polytope().unwrap().is_empty(), "Polytope is empty");
+                prop_assert!(star.is_empty(), "Empty star is not empty");
+                let result = panic::catch_unwind(|| {
+                    star.get_max(0)
+                });
+                prop_assert!(result.is_err(), "Infeasible star did not panic for get_min {:#?}", result);
+            }
+
+            #[test]
+            fn test_get_min_box_polytope(basis in array2(2, 2)) {
+                let num_dims = 2;
+                let box_bounds: Bounds1<f64> = Bounds1::new(Array1::from_elem(num_dims, -20.).view(), Array1::from_elem(num_dims, -20.).view());
+                let box_ineq = Inequality::new(Array2::zeros([1, num_dims]), Array1::zeros(1), box_bounds);
+                let poly = Polytope::from_halfspaces(box_ineq);
+
+                let center = arr1(&[0.0, 0.0]);
+                let star = Star2::new(basis, center).with_constraints(poly);
+
+                let result = panic::catch_unwind(|| {
+                    star.get_min(0);
+                });
+                assert!(result.is_ok());
+            }
+
+            #[test]
+            fn test_get_max_box_polytope(basis in array2(2, 2)) {
+                let num_dims = 2;
+                let box_bounds: Bounds1<f64> = Bounds1::new(Array1::from_elem(num_dims, -20.).view(), Array1::from_elem(num_dims, -20.).view());
+                let box_ineq = Inequality::new(Array2::zeros([1, num_dims]), Array1::zeros(1), box_bounds);
+                let poly = Polytope::from_halfspaces(box_ineq);
+
+                let center = arr1(&[0.0, 0.0]);
+                let star = Star2::new(basis, center).with_constraints(poly);
+
+                let result = panic::catch_unwind(|| {
+                    star.get_max(0);
+                });
+                assert!(result.is_ok());
+            }
         }
-        */
-
-        #[test]
-        fn test_get_min_feasible(star in non_empty_star(2,3)) {
-            prop_assert!(!star.input_space_polytope().unwrap().is_empty(), "Polytope is empty");
-            prop_assert!(!star.is_empty(), "Non empty star is empty");
-            let result = panic::catch_unwind(|| {
-                star.get_min(0);
-            });
-            prop_assert!(result.is_ok(), "Calculating min resulted in panic for feasible star");
-        }
-
-        #[test]
-        fn test_get_min_infeasible(star in empty_star(2,1)) {
-            prop_assert!(star.input_space_polytope().unwrap().is_empty(), "Polytope is empty");
-            prop_assert!(star.is_empty(), "Empty star is not empty");
-            let result = panic::catch_unwind(|| {
-                star.get_min(0)
-            });
-            prop_assert!(result.is_err(), "Infeasible star did not panic for get_min {:#?}", result);
-        }
-
-        #[test]
-        fn test_get_max_feasible(star in non_empty_star(2,3)) {
-            prop_assert!(!star.input_space_polytope().unwrap().is_empty(), "Polytope is empty");
-            prop_assert!(!star.is_empty(), "Non empty star is empty");
-            let result = panic::catch_unwind(|| {
-                star.get_max(0);
-            });
-            prop_assert!(result.is_ok(), "Calculating min resulted in panic for feasible star");
-        }
-
-        #[test]
-        fn test_get_max_infeasible(star in empty_star(2,1)) {
-            prop_assert!(star.input_space_polytope().unwrap().is_empty(), "Polytope is empty");
-            prop_assert!(star.is_empty(), "Empty star is not empty");
-            let result = panic::catch_unwind(|| {
-                star.get_max(0)
-            });
-            prop_assert!(result.is_err(), "Infeasible star did not panic for get_min {:#?}", result);
-        }
-
-        #[test]
-        fn test_get_min_box_polytope(basis in array2(2, 2)) {
-            let num_dims = 2;
-            let box_coeffs = Array2::eye(num_dims);
-            let mut box_rhs = Array1::ones(num_dims);
-            box_rhs *= 20.;
-
-            let mut box_ineq = Inequality::new(box_coeffs.clone(), box_rhs.clone());
-            let lower_box_ineq = Inequality::new(-1. * box_coeffs, box_rhs);
-
-            box_ineq.add_eqns(&lower_box_ineq, true);
-            let poly = Polytope::from_halfspaces(box_ineq);
-
-            let center = arr1(&[0.0, 0.0]);
-            let star = Star2::new(basis, center).with_constraints(poly);
-
-            let result = panic::catch_unwind(|| {
-                star.get_min(0);
-            });
-            assert!(result.is_ok());
-        }
-
-        #[test]
-        fn test_get_max_box_polytope(basis in array2(2, 2)) {
-            let num_dims = 2;
-            let box_coeffs = Array2::eye(num_dims);
-            let mut box_rhs = Array1::ones(num_dims);
-            box_rhs *= 20.;
-
-            let mut box_ineq = Inequality::new(box_coeffs.clone(), box_rhs.clone());
-            let lower_box_ineq = Inequality::new(-1. * box_coeffs, box_rhs);
-
-            box_ineq.add_eqns(&lower_box_ineq, true);
-            let poly = Polytope::from_halfspaces(box_ineq);
-
-            let center = arr1(&[0.0, 0.0]);
-            let star = Star2::new(basis, center).with_constraints(poly);
-
-            let result = panic::catch_unwind(|| {
-                star.get_max(0);
-            });
-            assert!(result.is_ok());
-        }
-
-
-    }
+    */
 }
